<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.0">
  
  <!-- This file is part of xgridfit.
       It is distributed under the GNU Public License, version 2.
       Copyright (c) 2006-8 by Peter S. Baker
  -->
  
  <xsl:template name="get-highest-function-number">
    <xsl:param name="current-function"/>
    <xsl:param name="current-max" select="0"/>
    <xsl:variable name="new-max">
      <xsl:choose>
        <xsl:when test="number($current-max) &gt;
			number($current-function/@num)">
          <xsl:value-of select="$current-max"/>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="$current-function/@num"/>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:choose>
      <xsl:when test="$current-function/following-sibling::function[@num]">
        <xsl:call-template name="get-highest-function-number">
          <xsl:with-param name="current-function"
            select="$current-function/following-sibling::function[@num][1]"/>
          <xsl:with-param name="current-max" select="$new-max"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="$new-max"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>
  
  <xsl:template name="get-function-number">
    <xsl:param name="function-name"/>
    <xsl:if test="not(key('function-index',$function-name))">
      <xsl:call-template name="error-message">
	<xsl:with-param name="msg">
	  <xsl:text>Function </xsl:text>
	  <xsl:value-of select="$function-name"/>
	  <xsl:text> not found</xsl:text>
	</xsl:with-param>
      </xsl:call-template>
    </xsl:if>
    <xsl:choose>
      <xsl:when test="key('function-index',$function-name)/@num">
	<xsl:value-of select="key('function-index',$function-name)/@num"/>
      </xsl:when>
      <xsl:otherwise>
	<xsl:value-of select="number($auto-function-base) +
			      number($predefined-functions) +
			      count(key('function-index',
			      $function-name)/preceding-sibling::function[not(@num)])"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>
  
  <!--
      This is called from a <call-function> instruction, but it
      recursively surveys the <param> elements in the <function> being
      called, counting up those params that can be resolved to numbers
      at compile time. These are "simple" params. Those that must be
      resolved at run time are not simple. This returns the count of
      simple params, followed by an upright, followed by a size byte:
      B if all are between 0 and 255, W if some are outside that
      range, A if some numbers aren't simple.
  -->
  <xsl:template match="param" mode="count-simple-params">
    <!-- A container for <with-param> elements. -->
    <xsl:param name="wps"/>
    <!-- The glyph (if any) that is an ancestor of the <call-function>. -->
    <xsl:param name="gl"/>
    <!-- Our running count (initialized to 0). -->
    <xsl:param name="count" select="0"/>
    <xsl:param name="size" select="'B'"/>
    <xsl:param name="call-macro-param-set"/>
    <xsl:variable name="nm" select="@name"/>
    <xsl:variable name="call-value">
      <xsl:choose>
	<xsl:when test="$wps/with-param[@name=$nm or @param-id=$nm]">
	  <xsl:call-template name="get-number-literal">
	    <xsl:with-param name="val"
			    select="$wps/with-param[@name=$nm or @param-id=$nm]/@value"/>
	    <xsl:with-param name="offset"
			    select="$wps/with-param[@name=$nm or @param-id=$nm]/@offset"/>
	    <xsl:with-param name="ancestor-glyph" select="$gl"/>
	    <xsl:with-param name="permitted" select="'1xrnc'"/>
	    <xsl:with-param name="call-macro-param-set" select="$call-macro-param-set"/>
	  </xsl:call-template>
	</xsl:when>
	<xsl:when test="@value">
	  <xsl:call-template name="get-number-literal">
	    <xsl:with-param name="val" select="@value"/>
	    <xsl:with-param name="permitted" select="'1xrnc'"/>
	    <xsl:with-param name="call-macro-param-set" select="$call-macro-param-set"/>
	  </xsl:call-template>
	</xsl:when>
	<xsl:otherwise>
	  <!-- If this happens, there will be an error later. -->
	  <xsl:text>NaN</xsl:text>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:variable name="new-count">
      <xsl:choose>
	<xsl:when test="number($call-value) or number($call-value) = 0">
	  <xsl:value-of select="number($count) + 1"/>
	</xsl:when>
	<xsl:otherwise>
	  <xsl:value-of select="$count"/>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:variable name="new-size">
      <xsl:choose>
	<xsl:when test="number($call-value) &lt; 0 or
			number($call-value) &gt; 255 or
			$size = 'W'">
	  <xsl:value-of select="'W'"/>
	</xsl:when>
	<xsl:when test="number($call-value) &gt;= 0 and
			number($call-value) &lt;= 255 and
			$size = 'B'">
	  <xsl:value-of select="'B'"/>
	</xsl:when>
	<xsl:otherwise>
	  <xsl:value-of select="'A'"/>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <!-- Are we done, or shall we recurse? -->
    <xsl:choose>
      <xsl:when test="following-sibling::param">
        <xsl:apply-templates select="following-sibling::param[1]"
                             mode="count-simple-params">
          <xsl:with-param name="wps" select="$wps"/>
          <xsl:with-param name="gl" select="$gl"/>
          <xsl:with-param name="count" select="$new-count"/>
	  <xsl:with-param name="size" select="$new-size"/>
	  <xsl:with-param name="call-macro-param-set" select="$call-macro-param-set"/>
        </xsl:apply-templates>
      </xsl:when>
      <xsl:otherwise>
        <xsl:value-of select="concat($new-count,'|',$new-size)"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>
  
  <!-- The <param-set> template pushes a set of function parameters onto the
       stack in preparation for a function call. Various complications: we
       determine what the parameters are and their proper order by looking
       into the function definition itself, so we lose the present context.
       Also, we'd like to optimize if possible by pushing all the parameters
       with a single PUSH command. Finally, we may omit a <with-param> element
       from the <param-set> if the corresponding <param> in the function
       definition has a default value; so we need to be ready to look for
       that default if the <with-param> is omitted here. -->
  <xsl:template match="param-set|call-function" mode="function">
    <xsl:param name="call-macro-param-set"/>
    <xsl:param name="f">
      <xsl:choose>
	<xsl:when test="ancestor-or-self::call-function/@name">
	  <xsl:value-of select="ancestor-or-self::call-function/@name"/>
	</xsl:when>
	<xsl:otherwise>
	  <xsl:value-of select="ancestor-or-self::call-function/@function-id"/>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:param>
    <xsl:param name="all-params" select="key('function-index',$f)/param |
					 key('function-index',$f)/params/param"/>
    <xsl:if test="local-name() = 'param-set'">
      <xsl:call-template name="debug-start"/>
    </xsl:if>
    <!-- Check the with-param elements for validity, since the schema does little -->
    <xsl:apply-templates select="with-param" mode="check-func"/>
    <!-- These variables allow us to carry some stuff from this
         context along when the for-each loop changes the context. -->
    <xsl:variable name="wps" select="."/>
    <xsl:variable name="ancestor-glyph" select="ancestor::glyph"/>
    <xsl:variable name="ancestor-function" select="ancestor::function"/>
    <xsl:variable name="ancestor-prep" select="ancestor::pre-program"/>
    <xsl:variable name="p">
      <xsl:choose>
        <xsl:when test="ancestor::function">
          <xsl:text>1xmpfvnc</xsl:text>
        </xsl:when>
        <xsl:otherwise>
          <xsl:text>1xmpvnc</xsl:text>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:variable name="csp-output">
      <xsl:apply-templates select="$all-params[1]"
			   mode="count-simple-params">
	<xsl:with-param name="wps" select="."/>
	<xsl:with-param name="gl" select="ancestor::glyph"/>
	<xsl:with-param name="call-macro-param-set" select="$call-macro-param-set"/>
      </xsl:apply-templates>
    </xsl:variable>
    <xsl:variable name="simple-param-count" select="substring-before($csp-output,'|')"/>
    <xsl:variable name="simple-param-size" select="substring-after($csp-output,'|')"/>
    <!-- Is this clear enough? This variable is true when we CAN optimize by
         pushing all the parameters with a single PUSHB command. The main test
         is that the number of "simple" params (those that can be resolved to
         numbers between 0 and 255 at compile time) is equal to the number of
         paramters to push. This optimiztion can be forced by the attribute
         opt="yes" or inhibited by the attribute opt="no" but is not influenced
         by the absence of an opt attribute. -->
    <xsl:variable name="push-all-at-once"
                  select="boolean((@opt = 'yes' or
                          number($simple-param-count) =
                          count($all-params)) and
                          (not(@opt) or @opt != 'no'))"/>
    <xsl:if test="$push-all-at-once and ($simple-param-size = 'A')">
      <xsl:call-template name="error-message">
	<xsl:with-param name="msg">
	  <xsl:text>Internal error in function call: if push-all-at-once is true,
a size must be given.</xsl:text>
	</xsl:with-param>
      </xsl:call-template>
    </xsl:if>
    <!-- We're going to loop backwards through the params in the function element.
         As we go we try to match up with-param elements in the present param-set
         with the param elements in the function. It is an error if any are
         missing. -->
    <xsl:for-each select="$all-params">
      <xsl:sort select="position()" order="descending"/>
      <!--
	  $use-add-mode: do we want the PUSH instruction output for
	  this iteration of the loop?
      -->
      <xsl:variable name="use-add-mode" select="boolean($push-all-at-once and
						following-sibling::param)"/>
      <!--
	  $nm: @name in a variable, since we're going to lose the
	  context.
      -->
      <xsl:variable name="nm" select="@name"/>
      <!--
	  $val-to-push: the value we're going to push.
      -->
      <xsl:variable name="val-to-push">
	<xsl:choose>
	  <xsl:when test="$wps/with-param[@name=$nm or @param-id=$nm]">
	    <xsl:value-of select="$wps/with-param[@name=$nm or @param-id=$nm]/@value"/>
	  </xsl:when>
	  <xsl:when test="@value">
	    <xsl:value-of select="@value"/>
	  </xsl:when>
	  <xsl:otherwise>
	    <!-- Oops! A parameter is missing. -->
	    <xsl:call-template name="error-message">
	      <xsl:with-param name="msg">
		<xsl:text>Parameter "</xsl:text>
		<xsl:value-of select="current()/@name"/>
		<xsl:text>" missing in call to function "</xsl:text>
		<xsl:value-of select="$f"/>
		<xsl:text>"</xsl:text>
	      </xsl:with-param>
	    </xsl:call-template>
	  </xsl:otherwise>
	</xsl:choose>
      </xsl:variable>
      <!--
	  $offset: the offset, if any.
      -->
      <xsl:variable name="offset">
	<xsl:choose>
	  <xsl:when test="$wps/with-param[@name=$nm or @param-id=$nm]/@offset">
	    <xsl:value-of select="$wps/with-param[@name=$nm or @param-id=$nm]/@offset"/>
	  </xsl:when>
	  <xsl:otherwise>
	    <xsl:value-of select="0"/>
	  </xsl:otherwise>
	</xsl:choose>
      </xsl:variable>
      <xsl:choose>
	<xsl:when test="$push-all-at-once">
	  <xsl:variable name="v">
	    <xsl:call-template name="get-number-literal">
	      <xsl:with-param name="val" select="$val-to-push"/>
	      <xsl:with-param name="offset" select="$offset"/>
	      <xsl:with-param name="permitted" select="$p"/>
	      <xsl:with-param name="ancestor-glyph" select="$ancestor-glyph"/>
	      <xsl:with-param name="call-macro-param-set"
			      select="$call-macro-param-set"/>
	    </xsl:call-template>
	  </xsl:variable>
	  <xsl:call-template name="push-num">
	    <xsl:with-param name="num" select="$v"/>
	    <xsl:with-param name="expect" select="$simple-param-count"/>
	    <xsl:with-param name="add-mode" select="$use-add-mode"/>
	    <xsl:with-param name="size" select="$simple-param-size"/>
	  </xsl:call-template>
	</xsl:when>
	<xsl:otherwise>
	  <xsl:call-template name="push-value">
	    <xsl:with-param name="val" select="$val-to-push"/>
            <xsl:with-param name="ancestor-glyph" select="$ancestor-glyph"/>
            <xsl:with-param name="ancestor-function" select="$ancestor-function"/>
            <xsl:with-param name="ancestor-prep" select="$ancestor-prep"/>
            <xsl:with-param name="offset" select="$offset"/>
            <xsl:with-param name="permitted" select="$p"/>
	    <xsl:with-param name="call-macro-param-set"
			    select="$call-macro-param-set"/>
          </xsl:call-template>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:for-each>
    <xsl:if test="local-name() = 'param-set'">
      <xsl:call-template name="debug-end"/>
    </xsl:if>
  </xsl:template>

  <xsl:template match="call-function">
    <xsl:param name="call-macro-param-set"/>
    <xsl:call-template name="debug-start"/>
    <xsl:variable name="fid">
      <xsl:choose>
	<xsl:when test="@name">
	  <xsl:value-of select="@name"/>
	</xsl:when>
	<xsl:when test="@function-id">
	  <xsl:value-of select="@function-id"/>
	</xsl:when>
	<xsl:otherwise>
	  <xsl:call-template name="error-message">
	    <xsl:with-param name="msg">
	      <xsl:text>Encountered &lt;call-function&gt; element without name or function-id</xsl:text>
	    </xsl:with-param>
	  </xsl:call-template>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:variable name="function-has-variables"
		  select="key('function-index',$fid)/variable or
			  key('function-index',$fid)/variables/variable"/>
    <!-- If we're calling this from a function, we need to preserve
         the variable $var-function-stack-count and restore it after
         we're done. -->
    <xsl:if test="ancestor::function">
      <xsl:call-template name="number-command">
        <xsl:with-param name="num" select="$var-function-stack-count"/>
        <xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
    </xsl:if>
    <!--
	If the function we're calling uses variables, we need to save
	the current variable frame.
    -->
    <xsl:if test="$function-has-variables">
      <xsl:call-template name="number-command">
	<xsl:with-param name="num" select="$var-frame-bottom"/>
	<xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
    </xsl:if>
    <!-- push the parameters onto the stack. -->
    <xsl:choose>
      <xsl:when test="param-set">
	<xsl:apply-templates select="param-set" mode="function">
	  <xsl:with-param name="call-macro-param-set"
			  select="$call-macro-param-set"/>
	</xsl:apply-templates>
      </xsl:when>
      <xsl:otherwise>
	<xsl:apply-templates select="." mode="function">
	  <xsl:with-param name="call-macro-param-set"
			  select="$call-macro-param-set"/>
	</xsl:apply-templates>
      </xsl:otherwise>
    </xsl:choose>
    <!-- continue saving variable frame. -->
    <xsl:if test="$function-has-variables">
      <xsl:call-template name="push-num">
        <xsl:with-param name="num" select="$var-frame-bottom"/>
	<xsl:with-param name="expect" select="2"/>
      </xsl:call-template>
      <xsl:call-template name="push-num">
	<xsl:with-param name="num" select="$var-frame-top"/>
	<xsl:with-param name="add-mode" select="true()"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'WS'"/>
      </xsl:call-template>
    </xsl:if>
    <!-- Push the index of this function. This always matches the number that
         was used to define the function when the fgpm table was set up. -->
    <xsl:call-template name="push-num">
      <xsl:with-param name="num">
        <xsl:call-template name="get-function-number">
          <xsl:with-param name="function-name" select="$fid"/>
        </xsl:call-template>
      </xsl:with-param>
    </xsl:call-template>
    <!-- Call via CALL or LOOPCALL, depending on whether we have two sets of
         parameters, or more, or fewer. -->
    <xsl:choose>
      <xsl:when test="count(param-set) &gt; 1">
        <xsl:call-template name="push-num">
          <xsl:with-param name="num" select="count(param-set)"/>
        </xsl:call-template>
        <xsl:call-template name="simple-command">
          <xsl:with-param name="cmd" select="'SWAP'"/>
        </xsl:call-template>
        <xsl:call-template name="simple-command">
          <xsl:with-param name="cmd" select="'LOOPCALL'"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:call-template name="simple-command">
          <xsl:with-param name="cmd" select="'CALL'"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
    <!-- Now restore variable frame. Copy var-frame-bottom to var-frame-top
         and then copy the number we saved to the stack before the call into
         var-frame-bottom. -->
    <xsl:if test="$function-has-variables">
      <xsl:call-template name="push-num">
        <xsl:with-param name="num" select="$var-frame-top"/>
	<xsl:with-param name="expect" select="2"/>
      </xsl:call-template>
      <xsl:call-template name="push-num">
        <xsl:with-param name="num" select="$var-frame-bottom"/>
	<xsl:with-param name="add-mode" select="true()"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'WS'"/>
      </xsl:call-template>
      <xsl:call-template name="stack-top-to-storage">
        <xsl:with-param name="loc" select="$var-frame-bottom"/>
      </xsl:call-template>
    </xsl:if>
    <!-- If we're calling this from a function, we need to restore
         the previous value of $var-function-stack-count (it's supposed
         to be on top of the stack after the function returns). -->
    <xsl:if test="ancestor::function">
      <xsl:call-template name="stack-top-to-storage">
        <xsl:with-param name="loc" select="$var-function-stack-count"/>
      </xsl:call-template>
    </xsl:if>
    <!-- If the function we are calling has the attribute return="yes" then
         we collect the return value from var-return-value. We'd like to
         store it in a variable (or whatever) referenced by "result-to", but
         if that attribute is missing we issue a warning and leave the value
         on the stack. -->
    <xsl:choose>
      <xsl:when test="key('function-index',$fid)/@return = 'yes'">
        <xsl:choose>
          <xsl:when test="@result-to">
            <xsl:call-template name="number-command">
              <xsl:with-param name="num" select="$var-return-value"/>
              <xsl:with-param name="cmd" select="'RS'"/>
            </xsl:call-template>
            <xsl:call-template name="store-value">
              <xsl:with-param name="vname" select="@result-to"/>
	      <xsl:with-param name="call-macro-param-set"
			      select="$call-macro-param-set"/>
            </xsl:call-template>
          </xsl:when>
          <xsl:otherwise>
            <xsl:call-template name="warning">
              <xsl:with-param name="msg">
                <xsl:text>Function </xsl:text>
                <xsl:value-of select="$fid"/>
                <xsl:text> returns a value, but I can</xsl:text>
                <xsl:value-of select="$newline"/>
                <xsl:text>find no place to put it (no "result-to" attribute).</xsl:text>
                <xsl:value-of select="$newline"/>
                <xsl:text>It is being left on the stack.</xsl:text>
              </xsl:with-param>
            </xsl:call-template>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>
        <xsl:if test="@result-to">
          <xsl:call-template name="warning">
            <xsl:with-param name="msg">
              <xsl:text>The return value of a function cannot be read </xsl:text>
              <xsl:text>unless the</xsl:text>
              <xsl:value-of select="$newline"/>
              <xsl:text>attribute return="yes" is present. </xsl:text>
              <xsl:text>I am ignoring the "result-to"</xsl:text>
              <xsl:value-of select="$newline"/>
              <xsl:text>attribute of this call to function </xsl:text>
              <xsl:value-of select="$fid"/>
              <xsl:text>.</xsl:text>
            </xsl:with-param>
          </xsl:call-template>
        </xsl:if>
      </xsl:otherwise>
    </xsl:choose>
    <xsl:call-template name="debug-end"/>
  </xsl:template>

  <xsl:template match="function">
    <xsl:call-template name="debug-start"/>
    <xsl:variable name="omit-initial-newline-i">
      <xsl:choose>
        <xsl:when test="@num and not(preceding-sibling::function[@num])">
          <xsl:text>1</xsl:text>
        </xsl:when>
        <xsl:otherwise>
          <xsl:text>0</xsl:text>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:variable name="omit-initial-newline" select="boolean(number($omit-initial-newline-i))"/>
    <xsl:call-template name="number-command">
      <xsl:with-param name="num">
        <xsl:call-template name="get-function-number">
          <xsl:with-param name="function-name">
	    <xsl:choose>
	      <xsl:when test="@name">
		<xsl:value-of select="@name"/>
	      </xsl:when>
	      <xsl:when test="@id">
		<xsl:value-of select="@id"/>
	      </xsl:when>
	      <xsl:otherwise>
		<xsl:call-template name="error-message">
		  <xsl:with-param name="msg">
		    <xsl:text>Encountered &lt;function&lt; lacking name and id attributes.</xsl:text>
		  </xsl:with-param>
		</xsl:call-template>
	      </xsl:otherwise>
	    </xsl:choose>
	  </xsl:with-param>
        </xsl:call-template>
      </xsl:with-param>
      <xsl:with-param name="cmd" select="'FDEF'"/>
      <xsl:with-param name="with-leading-newline" select="not($omit-initial-newline)"/>
    </xsl:call-template>
    <!-- First save the current depth of the stack in the Storage Location
         var-function-stack-count. This will help us find
         parameters when we need them, even if stuff has been pushed onto the
         stack in the meantime. -->
    <xsl:if test="param|params/param">
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'DEPTH'"/>
      </xsl:call-template>
      <xsl:call-template name="stack-top-to-storage">
        <xsl:with-param name="loc" select="$var-function-stack-count"/>
      </xsl:call-template>        
    </xsl:if>
    <!-- If we have declared variables, place the new top of the variable
         frame in var-frame-top. -->
    <xsl:if test="variable or variables/variable">
      <xsl:call-template name="push-num">
        <xsl:with-param name="num" select="$var-frame-top"/>
      </xsl:call-template>
      <xsl:call-template name="number-command">
        <xsl:with-param name="num" select="$var-frame-bottom"/>
        <xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
      <xsl:call-template name="number-command">
        <xsl:with-param name="num" select="count(variable|variables/variable)"/>
        <xsl:with-param name="cmd" select="'ADD'"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'WS'"/>
      </xsl:call-template>
      <!-- Also, initialize any that want initializing. -->
      <xsl:apply-templates select="variable|variables/variable" mode="initialize"/>
    </xsl:if>
    <!-- If the function is declared as expecting a return value, initialize
         var-return-value to zero. This is a safety measure, since it is
         impractical (right now at least) to check that the function contains
         an instruction that writes to that location. -->
    <xsl:if test="@return = 'yes'">
      <xsl:call-template name="push-num">
        <xsl:with-param name="num" select="$var-return-value"/>
        <xsl:with-param name="expect" select="2"/>
      </xsl:call-template>
      <xsl:call-template name="push-num">
        <xsl:with-param name="num" select="0"/>
        <xsl:with-param name="add-mode" select="true()"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'WS'"/>
      </xsl:call-template>
    </xsl:if>
    <!-- Now execute instructions. -->
    <xsl:apply-templates/>
    <!-- Pop all the parameters off the stack, and we're done. -->
    <xsl:for-each select="param|params/param">
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'POP'"/>
      </xsl:call-template>
    </xsl:for-each>
    <xsl:call-template name="simple-command">
      <xsl:with-param name="cmd" select="'ENDF'"/>
    </xsl:call-template>
    <xsl:call-template name="debug-end"/>
  </xsl:template>

  <xsl:template match="legacy-functions">
    <xsl:call-template name="debug-start"/>
    <xsl:for-each select="*">
      <xsl:apply-templates select=".">
	<xsl:with-param name="with-leading-newline"
			select="boolean(preceding-sibling::*)"/>
      </xsl:apply-templates>
    </xsl:for-each>
    <xsl:call-template name="debug-end"/>
  </xsl:template>

  <xsl:template match="macro">
    <xsl:param name="call-macro-param-set"/>
    <xsl:call-template name="debug-start"/>
    <xsl:apply-templates>
      <xsl:with-param name="call-macro-param-set"
		      select="$call-macro-param-set"/>
    </xsl:apply-templates>
    <xsl:call-template name="debug-end"/>
  </xsl:template>

  <xsl:template match="with-param" mode="check-macro">
    <xsl:if test="not(@value) and not(set) and not(range) and not(line)">
      <xsl:call-template name="error-message">
	<xsl:with-param name="msg">
	  <xsl:text>In call to macro "</xsl:text>
	  <xsl:choose>
	    <xsl:when test="@name">
	      <xsl:value-of select="@name"/>
	    </xsl:when>
	    <xsl:when test="@macro-id">
	      <xsl:value-of select="@macro-id"/>
	    </xsl:when>
	  </xsl:choose>
	  <xsl:text>" &lt;with-param&gt; element "</xsl:text>
	  <xsl:choose>
	    <xsl:when test="@name">
	      <xsl:value-of select="@name"/>
	    </xsl:when>
	    <xsl:otherwise>
	      <xsl:value-of select="@param-id"/>
	    </xsl:otherwise>
	  </xsl:choose>
	  <xsl:text>" must have a value attribute or contain &lt;line&gt;, &lt;set&gt; or &lt;range&gt;</xsl:text>
	</xsl:with-param>
      </xsl:call-template>
    </xsl:if>
  </xsl:template>

  <xsl:template match="with-param" mode="check-func">
    <xsl:variable name="fid">
      <xsl:choose>
	<xsl:when test="ancestor::call-function/@name">
	  <xsl:value-of select="ancestor::call-function/@name"/>
	</xsl:when>
	<xsl:otherwise>
	  <xsl:value-of select="ancestor::call-function/@function-id"/>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:if test="not(@value)">
      <xsl:call-template name="error-message">
	<xsl:with-param name="msg">
	  <xsl:text>In call to function "</xsl:text>
	  <xsl:value-of select="$fid"/>
	  <xsl:text>" &lt;with-param&gt; element "</xsl:text>
	  <xsl:choose>
	    <xsl:when test="@name">
	      <xsl:value-of select="@name"/>
	    </xsl:when>
	    <xsl:otherwise>
	      <xsl:value-of select="@param-id"/>
	    </xsl:otherwise>
	  </xsl:choose>
	  <xsl:text>" must have a value attribute</xsl:text>
	</xsl:with-param>
      </xsl:call-template>
    </xsl:if>
  </xsl:template>

  <xsl:template match="param-set" mode="macro">
    <xsl:param name="mid"/>
    <xsl:apply-templates select="with-param" mode="check-macro"/>
    <xsl:apply-templates select="key('macro-index',$mid)">
      <xsl:with-param name="call-macro-param-set" select="with-param"/>
    </xsl:apply-templates>
  </xsl:template>

  <xsl:template match="params"></xsl:template>

  <xsl:template match="call-macro">
    <xsl:call-template name="debug-start"/>
    <xsl:variable name="mid">
      <xsl:choose>
	<xsl:when test="@name">
	  <xsl:value-of select="@name"/>
	</xsl:when>
	<xsl:when test="@macro-id">
	  <xsl:value-of select="@macro-id"/>
	</xsl:when>
	<xsl:otherwise>
	  <xsl:call-template name="error-message">
	    <xsl:with-param name="msg">
	      <xsl:text>Encountered a &lt;call-macro&gt; without name or macro-id.</xsl:text>
	    </xsl:with-param>
	  </xsl:call-template>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <xsl:if test="not(key('macro-index',$mid))">
      <xsl:call-template name="error-message">
	<xsl:with-param name="msg">
	  <xsl:text>Can't find macro </xsl:text>
	  <xsl:value-of select="$mid"/>
	</xsl:with-param>
      </xsl:call-template>
    </xsl:if>
    <xsl:choose>
      <xsl:when test="param-set">
	<xsl:apply-templates select="param-set" mode="macro">
	  <xsl:with-param name="mid" select="$mid"/>
	</xsl:apply-templates>
      </xsl:when>
      <xsl:when test="with-param">
	<xsl:apply-templates select="with-param" mode="check-macro"/>
	<xsl:apply-templates select="key('macro-index',$mid)">
	  <xsl:with-param name="call-macro-param-set" select="with-param"/>
	</xsl:apply-templates>
      </xsl:when>
      <xsl:otherwise>
	<xsl:apply-templates select="key('macro-index',$mid)"/>
      </xsl:otherwise>
    </xsl:choose>
    <xsl:call-template name="debug-end"/>
  </xsl:template>

  <xsl:template match="call-glyph">
    <xsl:call-template name="debug-start"/>
    <xsl:variable name="n" select="@ps-name"/>
    <xsl:if test="not(/xgridfit/glyph[@ps-name=$n])">
      <xsl:call-template name="error-message">
	<xsl:with-param name="msg">
	  <xsl:text>Cannot find glyph program with psname="</xsl:text>
	  <xsl:value-of select="@ps-name"/>
	  <xsl:text>"</xsl:text>
	</xsl:with-param>
      </xsl:call-template>
    </xsl:if>
    <xsl:variable name="save-variable-frame"
		  select="key('glyph-index',$n)/variable or
			  key('glyph-index',$n)/variables/variable"/>
    <!-- Save the variable frame -->
    <xsl:if test="$save-variable-frame">
      <xsl:call-template name="push-num">
	<xsl:with-param name="num" select="$var-frame-bottom"/>
	<xsl:with-param name="expect" select="2"/>
      </xsl:call-template>
      <xsl:call-template name="push-num">
	<xsl:with-param name="num" select="$var-frame-bottom"/>
	<xsl:with-param name="add-mode" select="true()"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
	<xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
      <xsl:call-template name="number-command">
	<xsl:with-param name="num" select="$var-frame-top"/>
	<xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
	<xsl:with-param name="cmd" select="'WS'"/>
      </xsl:call-template>
    </xsl:if>
    <!-- Call the glyph program, passing in any parameters -->
    <xsl:apply-templates select="key('glyph-index',$n)" mode="called">
      <xsl:with-param name="call-macro-param-set" select="with-param"/>
    </xsl:apply-templates>
    <!-- Restore the variable frame -->
    <xsl:if test="$save-variable-frame">
      <xsl:call-template name="push-num">
	<xsl:with-param name="num" select="$var-frame-top"/>
	<xsl:with-param name="expect" select="2"/>
      </xsl:call-template>
      <xsl:call-template name="push-num">
	<xsl:with-param name="num" select="$var-frame-bottom"/>
	<xsl:with-param name="add-mode" select="true()"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
	<xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
	<xsl:with-param name="cmd" select="'WS'"/>
      </xsl:call-template>
      <xsl:call-template name="stack-top-to-storage">
	<xsl:with-param name="loc" select="$var-frame-bottom"/>
      </xsl:call-template>
    </xsl:if>
    <xsl:call-template name="debug-end"/>
  </xsl:template>

  <xsl:template match="glyph" mode="called">
    <xsl:param name="call-macro-param-set"/>
    <xsl:if test="variable or variables/variable">
      <xsl:call-template name="push-num">
        <xsl:with-param name="num" select="$var-frame-top"/>
	<xsl:with-param name="expect" select="2"/>
      </xsl:call-template>
      <xsl:call-template name="push-num">
	<xsl:with-param name="num" select="$var-frame-bottom"/>
	<xsl:with-param name="add-mode" select="true()"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'RS'"/>
      </xsl:call-template>
      <xsl:call-template name="number-command">
        <xsl:with-param name="num" select="count(variable|variables/variable)"/>
        <xsl:with-param name="cmd" select="'ADD'"/>
      </xsl:call-template>
      <xsl:call-template name="simple-command">
        <xsl:with-param name="cmd" select="'WS'"/>
      </xsl:call-template>
      <!-- Also, initialize any that want initializing. -->
      <xsl:apply-templates select="variable|variables/variable" mode="initialize"/>
    </xsl:if>
    <xsl:apply-templates>
      <xsl:with-param name="call-macro-param-set" select="$call-macro-param-set"/>
    </xsl:apply-templates>
  </xsl:template>

  <xsl:template match="glyph|function|macro" mode="survey-vars">
    <xsl:choose>
      <xsl:when test="variable|variables/variable">
	<xsl:text>v</xsl:text>
      </xsl:when>
      <xsl:otherwise>
	<xsl:apply-templates select="call-function|call-glyph|call-macro" mode="survey-vars"/>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="call-glyph|call-function|call-macro" mode="survey-vars">
    <xsl:variable name="l" select="local-name()"/>
    <xsl:choose>
      <xsl:when test="$l = 'call-glyph'">
	<xsl:text>g</xsl:text>
	<xsl:apply-templates select="key('glyph-index',@ps-name)" mode="survey-vars"/>
      </xsl:when>
      <xsl:when test="$l = 'call-function'">
	<xsl:text>f</xsl:text>
	<xsl:apply-templates select="key('function-index',@name)" mode="survey-vars"/>
      </xsl:when>
      <xsl:when test="$l = 'call-macro'">
	<xsl:text>m</xsl:text>
	<xsl:apply-templates select="key('macro-index',@name)" mode="survey-vars"/>
      </xsl:when>
    </xsl:choose>
  </xsl:template>

</xsl:stylesheet>
